import typing
from typing import Any
import numpy as np
import pandas as pd
import statsmodels.api as sm

from pier_ds_utils.transformer import BaseCustomTransformer
from pier_ds_utils.prep import add_constant_column
from sklearn.base import BaseEstimator


class GLMWrapper(BaseCustomTransformer):
    def __init__(
        self, add_constant: bool = True, os_factor: np.float64 = 1.0, **init_params
    ):
        self._add_constant = add_constant
        self.os_factor = os_factor
        self.init_params = init_params

    def get_params(self, deep=True):
        return {
            **self.init_params,
            **{"add_constant": self._add_constant, "os_factor": self.os_factor},
        }

    def fit(self, X, y, **fit_params):
        if self._add_constant:
            X = add_constant_column(
                X, prepend=True, constant_value=1.0, column_name="const"
            )

        self.model_ = sm.GLM(endog=y, exog=X, **self.init_params)
        fit_method = fit_params.pop("fit_method", "fit")
        self.results_ = getattr(self.model_, fit_method)(**fit_params)
        return self

    def predict(self, X, **predict_params):
        if self._add_constant:
            X = add_constant_column(
                X, prepend=True, constant_value=1.0, column_name="const"
            )

        return self.results_.predict(exog=X, **predict_params) * self.os_factor


class PredictProbaSelector(BaseCustomTransformer):
    def __init__(self, model: BaseEstimator, column: typing.Union[str, int] = None):
        self.model = model
        self.column = column

    def predict_proba(self, X: pd.DataFrame, **kwargs) -> np.ndarray:
        return self.model.predict_proba(X, **kwargs)[:, self.column].tolist()

    def fit(self, X: pd.DataFrame, y: pd.Series, **kwargs):
        return self.model.fit(X, y, **kwargs)

    def __getattr__(self, __name: str) -> Any:
        if not hasattr(super(), __name):
            return getattr(self.model, __name)

    def get_params(self, deep: bool = True) -> dict:
        return {
            "model": self.model.get_params(deep=deep) if deep else self.model,
            "column": self.column,
        }



import pandas as pd
import typing


def add_constant_column(
    df: typing.Union[pd.DataFrame, pd.Series],
    prepend: bool = True,
    column_name: str = "const",
    constant_value: typing.Any = 1.0,
) -> pd.DataFrame:
    """
    Add a constant column to a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to which the constant column will be added.
    prepend : bool, default=True
        If True, the constant column will be added as the first column.
        If False, the constant column will be added as the last column.
    column_name : str, default="const"
        The name of the constant column.
    constant_value : Any, default=1.0
        The value to be used for the constant column.

    Returns
    -------
    pd.DataFrame
        The DataFrame with the constant column added.
    """
    if not isinstance(df, pd.DataFrame):
        raise ValueError("df must be a DataFrame or Series")

    df = df.copy()

    original_column_names = df.columns.tolist()

    df[column_name] = constant_value

    if column_name in original_column_names:
        original_column_names.remove(column_name)

    column_names = (
        [column_name] + original_column_names
        if prepend
        else original_column_names + [column_name]
    )

    return df[column_names]



import pandas as pd
import numpy as np
import typing
from pier_ds_utils.transformer import BaseCustomTransformer
import functools


class StaticGLM(BaseCustomTransformer):
    def __init__(
        self,
        coefficients_map: typing.Dict[str, typing.Union[int, float]],
        constant: typing.Optional[typing.Union[int, float]] = 0.0,
        os_factor: typing.Optional[typing.Union[int, float]] = 1.0,
    ) -> None:
        """
        A static Generalized Linear Model (GLM) created from a dictionary of coefficients.
        The model is used for making predictions based on a linear combination of features.

        Parameters
        ----------
        coefficients_map : dict
            A dictionary mapping feature names to their corresponding coefficients.
        constant : float, optional
            A constant term added to the linear combination of features. Default is 0.0.
        os_factor : float, optional
            A scaling factor applied to the predictions. Default is 1.0.
        """
        super().__init__()
        self.coefficients_map = coefficients_map
        self.constant = constant
        self.os_factor = os_factor
        self.fitted_ = True

    @property
    def coefficients_map(self) -> typing.Dict[str, float]:
        """
        Get the coefficients map.
        """
        return self._coefficients_map

    @coefficients_map.setter
    def coefficients_map(self, value: typing.Dict[str, float]) -> None:
        """
        Set the coefficients map.
        """
        if not isinstance(value, dict):
            raise ValueError("coefficients_map must be a dictionary.")

        if not value:
            raise ValueError("coefficients_map cannot be empty.")

        if any(not isinstance(k, str) for k in value.keys()):
            raise ValueError("All keys in coefficients_map must be strings.")

        if any(k == "" for k in value.keys()):
            raise ValueError("All keys in coefficients_map must be non-empty strings.")

        self._coefficients_map = {}

        for key, val in value.items():
            if not isinstance(val, (int, float)):
                raise ValueError(
                    f"Value for {key} in coefficients_map must be a number."
                )
            self._coefficients_map[key] = float(val)

    @property
    def constant(self) -> float:
        """
        Get the constant.
        """
        return self._constant

    @constant.setter
    def constant(self, value: float) -> None:
        """
        Set the constant.
        """
        if not isinstance(value, (int, float)):
            raise ValueError("constant must be a number.")
        self._constant = float(value)

    @property
    def os_factor(self) -> float:
        """
        Get the os_factor.
        """
        return self._os_factor

    @os_factor.setter
    def os_factor(self, value: float) -> None:
        """
        Set the os_factor.
        """
        if not isinstance(value, (int, float)):
            raise ValueError("os_factor must be a number.")
        self._os_factor = float(value)

    @functools.cached_property
    def feature_names(self) -> typing.List[str]:
        """
        Get the feature names.
        """
        return list(self.coefficients_map.keys())

    @functools.cached_property
    def coefficients_(self) -> np.ndarray:
        """
        Get the coefficients.
        """
        return np.array(list(self.coefficients_map.values()), dtype=np.float64)

    def fit(self, X: pd.DataFrame, y: pd.Series) -> "StaticGLM":
        """
        Fit the model.
        """
        return self

    @staticmethod
    def _check_prediction_input(
        X: pd.DataFrame, required_cols: typing.List[str]
    ) -> None:
        """
        Check if the input DataFrame contains all required columns.
        """
        if not isinstance(X, pd.DataFrame):
            raise ValueError("Input must be a pandas DataFrame.")

        missing_cols = set(required_cols) - set(X.columns)
        if missing_cols:
            raise ValueError(
                f"Input DataFrame is missing the following columns: {missing_cols}"
            )

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        """
        Predict using the model.
        """
        self._check_prediction_input(X, self.feature_names)

        X = X.copy()

        # Ensure the DataFrame contains only the required columns
        X = X[self.feature_names]

        # Calculate the linear combination of features and coefficients
        linear_combination = np.dot(X.values, self.coefficients_)

        # Apply the os_factor and constant
        predictions = self.os_factor * (linear_combination + self.constant)

        return predictions



import typing

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class BaseCustomTransformer(BaseEstimator, TransformerMixin):
    def set_output(self, transform: str = "pandas") -> BaseEstimator:
        return self


class CustomDiscreteCategorizer(BaseCustomTransformer):
    def __init__(
        self,
        column: str,
        categories: typing.List[typing.List[typing.Any]],
        labels: typing.List[typing.Any],
        default_value: typing.Any = None,
        output_column: typing.Optional[str] = None,
    ):
        """
        Transformer to categorize a column into custom categories.

        Parameters
        ----------
        column: str
            Name of the column to be transformed
        categories: list of lists
            List of categories to be used for categorization. Each category must be a list of elements.
        labels: list
            List of labels to be used for categorization. Must have the same length as categories.
        default_value: any
            Value to be used for missing values. If None, missing values will be kept as NaN.
        output_column: str
            Name of the output column. If None, the original column will be overwritten.
        """
        if len(categories) != len(labels):
            raise ValueError(
                "Number of categories must be the same as number of labels"
            )

        for category in categories:
            if not isinstance(category, list):
                raise TypeError("Each category must be a list")

        self._column = column
        self._categories = categories
        self._labels = labels
        self._default_value = default_value
        self._output_column = output_column

    @property
    def categories_(self) -> typing.List[typing.List[typing.Any]]:
        return self._categories

    @property
    def labels_(self) -> typing.List[typing.Any]:
        return self._labels

    @classmethod
    def from_dict(cls, categories: typing.Dict, **kwargs):
        return cls(list(categories.values()), list(categories.keys()), **kwargs)

    def get_params(self, deep: bool = True) -> dict:
        return {
            "categories": self.categories_,
            "labels": self.labels_,
            "default_value": self._default_value,
            "output_column": self._output_column,
        }

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        values = X[self._column].copy()
        output = pd.Series(np.nan, index=X.index, dtype="object")

        for category, label in zip(self._categories, self._labels):
            output.loc[values.isin(category)] = label

        if self._default_value is not None:
            output.fillna(self._default_value, inplace=True)

        output_column = self._output_column or self._column
        X.loc[:, output_column] = output

        return X


class CustomIntervalCategorizer(BaseCustomTransformer):
    def __init__(
        self,
        column: str,
        intervals: typing.List[typing.Tuple[typing.Union[int, float]]],
        labels: typing.List[typing.Any],
        default_value: typing.Any = None,
        output_column: typing.Optional[str] = None,
    ):
        """
        Custom transformer to categorize a numeric column into intervals.

        Parameters
        ----------
        column: str
            Name of the column to be transformed
        intervals: list of tuples
            List of intervals to be used for categorization. Each interval must be a tuple with two elements.
            The first element must be smaller than the second. The comparison is inclusive for the first element (>=)
            and exclusive for the second (<).
        labels: list
            List of labels to be used for categorization. Must have the same length as intervals.
        default_value: any
            Value to be used for missing values. If None, missing values will be kept as NaN.
        output_column: str
            Name of the output column. If None, the original column will be overwritten.
        """
        if len(intervals) != len(labels):
            raise ValueError("Number of intervals must be the same as number of labels")

        for interval in intervals:
            if not isinstance(interval, tuple):
                raise TypeError("Each interval must be a tuple")

            if len(interval) != 2:
                raise ValueError("Each interval must have two elements")

            if not isinstance(interval[0], (int, float)) or not isinstance(
                interval[1], (int, float)
            ):
                raise TypeError("Each interval element must be a number")

            if interval[0] >= interval[1]:
                raise ValueError(
                    "Each interval must have the first element smaller than the second"
                )

        self._column = column
        self._intervals = intervals
        self._labels = labels
        self._default_value = default_value
        self._output_column = output_column

    @property
    def column_(self) -> str:
        return self._column

    @property
    def intervals_(self) -> typing.List[typing.Tuple[typing.Union[int, float]]]:
        return self._intervals

    @property
    def labels_(self) -> typing.List[typing.Any]:
        return self._labels

    @property
    def default_value_(self) -> typing.Any:
        return self._default_value

    @property
    def output_column_(self) -> str:
        return self._output_column

    def get_output_column(self) -> str:
        return self.output_column_ or self.column_

    @classmethod
    def from_dict(cls, intervals: typing.Dict, **kwargs):
        return cls(list(intervals.values()), list(intervals.keys()), **kwargs)

    def get_params(self, deep: bool = True) -> dict:
        return {
            "intervals": self.intervals_,
            "labels": self.labels_,
            "default_value": self.default_value_,
            "output_column": self.output_column_,
            "column": self.column_,
        }

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        values = X[self.column_].astype(float).copy()
        output = pd.Series(np.nan, index=X.index, dtype="object")

        for interval, label in zip(self.intervals_, self.labels_):
            output.loc[(values >= interval[0]) & (values < interval[1]),] = label

        if self.default_value_ is not None:
            output.fillna(self.default_value_, inplace=True)

        X.loc[:, self.get_output_column()] = output

        return X


class CustomIntervalCategorizerByCategory(BaseCustomTransformer):
    def __init__(
        self,
        category_column: str,
        interval_categorizers: typing.Dict[str, CustomIntervalCategorizer],
        default_categorizer: typing.Optional[CustomIntervalCategorizer] = None,
        default_value: typing.Any = None,
        output_column: typing.Optional[str] = None,
    ):
        """
        Custom transformer to categorize a numeric column into intervals given a categorical column.

        Parameters
        ----------
        category_column: str
            Name of the column to be used for categorization
        interval_categorizers: dict
            Dictionary of interval categorizers to be used for categorization. Keys must be the categories of the
            category_column and values must be CustomIntervalCategorizer.
        default_categorizer: CustomIntervalCategorizer
            Categorizer to be used when value does not match any categorizer in interval_categorizers.
        default_value: any
            Value to be used for missing values. If None, missing values will be kept as NaN.
        output_column: str
            Name of the output column. If None, the original column will be overwritten.
        """
        if not isinstance(interval_categorizers, dict):
            raise TypeError("interval_categorizers must be a dict")

        for key, value in interval_categorizers.items():
            if not isinstance(key, str):
                raise TypeError("Keys of interval_categorizers must be strings")

            if not isinstance(value, CustomIntervalCategorizer):
                raise TypeError(
                    "Values of interval_categorizers must be CustomIntervalCategorizer"
                )

        self._category_column = category_column
        self._interval_categorizers = interval_categorizers
        self._default_categorizer = default_categorizer
        self._default_value = default_value
        self._output_column = output_column

    @property
    def category_column_(self) -> str:
        return self._category_column

    @property
    def interval_categorizers_(self) -> typing.Dict[str, CustomIntervalCategorizer]:
        return self._interval_categorizers

    @property
    def default_categorizer_(self) -> typing.Optional[CustomIntervalCategorizer]:
        return self._default_categorizer

    @classmethod
    def from_dict(cls, **kwargs):
        return cls(**kwargs)

    def get_params(self, deep: bool = True) -> dict:
        return {
            "category_column": self.category_column_,
            "interval_categorizers": self.interval_categorizers_,
            "default_categorizer": self.default_categorizer_,
            "default_value": self._default_value,
            "output_column": self._output_column,
        }

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        output = pd.Series(np.nan, index=X.index, dtype="object")

        for category, interval_categorizer in self.interval_categorizers_.items():
            output.loc[
                X[self._category_column] == category
            ] = interval_categorizer.transform(
                X.loc[X[self._category_column] == category]
            )[interval_categorizer.get_output_column()]

        if self._default_categorizer is not None:
            output.loc[
                ~(X[self.category_column_].isin(self.interval_categorizers_.keys()))
            ] = self._default_categorizer.transform(
                X.loc[
                    ~(X[self.category_column_].isin(self.interval_categorizers_.keys()))
                ]
            )[self._default_categorizer.get_output_column()]

        if self._default_value is not None:
            output.fillna(self._default_value, inplace=True)

        output_column = self._output_column or self._category_column
        X.loc[:, output_column] = output

        return X


class LogTransformer(BaseCustomTransformer):
    """Calculates the natural logarithm of the input data. This transformer is useful for transforming skewed data into a more normal distribution."""

    def fit(self, X, y=None):
        # No fitting needed for this transformer
        return self

    def transform(self, X):
        # Apply log transformation (ensure values are positive for log)
        return X.apply(np.log, axis=1)


class BoundariesTransformer(BaseCustomTransformer):
    def __init__(self, lower_bound: float = None, upper_bound: float = None):
        """Transformer to apply lower and upper bounds to the input data.
        This transformer is useful for capping or flooring values in a dataset.
        The lower_bound and upper_bound parameters are optional. If not provided, the transformer will not apply any bounds.
        The bounds are inclusive, meaning that values equal to the bounds will be kept as is.
        """
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound

    def get_params(self, deep=True):
        return {
            "lower_bound": self._lower_bound,
            "upper_bound": self._upper_bound,
        }

    def fit(self, X, y=None):
        # No fitting needed for this transformer
        return self

    def transform(self, X):
        # Apply boundaries transformation
        if self._lower_bound is not None:
            X.clip(lower=self._lower_bound, inplace=True)
        if self._upper_bound is not None:
            X.clip(upper=self._upper_bound, inplace=True)
        return X
